"""
A module for testing pisek itself. The strategy is to take a functioning
fixture of a task and then break it in various small ways to see whether
pisek catches the problem.
"""

import unittest

from util import modify_config
from test_cms import TestSumCMS


class TestMissingTest(TestSumCMS):
    def expecting_success(self) -> bool:
        return False

    def modify_task(self) -> None:
        def modification_fn(raw_config):
            del raw_config["test01"]

        modify_config(self.task_dir, modification_fn)


class TestExtraConfigKeys(TestSumCMS):
    def expecting_success(self) -> bool:
        return False

    def modify_task(self) -> None:
        def modification_fn(raw_config):
            raw_config["task"]["foo"] = "bar"

        modify_config(self.task_dir, modification_fn)


class TestExtraConfigKeysInTest(TestSumCMS):
    def expecting_success(self) -> bool:
        return False

    def modify_task(self) -> None:
        def modification_fn(raw_config):
            raw_config["test01"]["foo"] = "bar"

        modify_config(self.task_dir, modification_fn)


class TestExtraConfigSection(TestSumCMS):
    def expecting_success(self) -> bool:
        return False

    def modify_task(self) -> None:
        def modification_fn(raw_config):
            raw_config.add_section("baz")
            raw_config["baz"]["foo"] = "bar"

        modify_config(self.task_dir, modification_fn)


if __name__ == "__main__":
    unittest.main()



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import abstractmethod
from decimal import Decimal
from typing import assert_never, Optional, override

from pisek.utils.text import tab
from pisek.utils.paths import IInputPath, IOutputPath, LogPath
from pisek.env.env import Env
from pisek.config.config_types import DataFormat
from pisek.task_jobs.tools import SanitizationResultKind
from pisek.jobs.jobs import State, PipelineItemFailure
from pisek.task_jobs.run_result import RunResult, RunResultKind
from pisek.task_jobs.program import ProgramsJob
from pisek.task_jobs.solution.solution_result import (
    Verdict,
    SolutionResult,
    RelativeSolutionResult,
)


class RunChecker(ProgramsJob):
    """Runs checker on single input. (Abstract class)"""

    def __init__(
        self,
        env: Env,
        name: str,
        test: int,
        checker_name: str,
        input_: IInputPath,
        checker_log_file: LogPath,
        expected_verdict: Optional[Verdict],
        **kwargs,
    ) -> None:
        super().__init__(env=env, name=name, **kwargs)
        self.test = test
        self.input = input_
        self.checker_name = checker_name
        self.checker_log_file = checker_log_file
        self.expected_verdict = expected_verdict

        self.checker_rr: RunResult | None = None
        self._register_cached_attribute("checker_rr")
        self.result: SolutionResult | None

    @abstractmethod
    def _get_solution_run_res_kind(self) -> RunResultKind:
        """
        Gets solution's RunResultKind.
        Might run the solution so call this only once!!!
        """
        pass

    @abstractmethod
    def _check(self) -> SolutionResult:
        """Here actually do the output checking."""
        pass

    @abstractmethod
    def _checking_message(self) -> str:
        pass

    def _checking_message_capitalized(self) -> str:
        msg = self._checking_message()
        return msg[0].upper() + msg[1:]

    def _run(self) -> SolutionResult:
        result = self._get_solution_result(self._get_solution_run_res_kind())

        if (
            self.expected_verdict is not None
            and result.verdict != self.expected_verdict
        ):
            raise PipelineItemFailure(
                f"{self._checking_message_capitalized()} should have got verdict '{self.expected_verdict}' but got '{result.verdict}'."
            )

        return result

    def _get_solution_result(self, kind: RunResultKind) -> SolutionResult:
        if kind == RunResultKind.OK:
            out_form = self._env.config.tests.out_format
            san_res = self.prerequisites_results.get("sanitize")
            if san_res is None:
                pass
            elif san_res.kind == SanitizationResultKind.invalid:
                return RelativeSolutionResult(
                    verdict=Verdict.normalization_fail,
                    message=san_res.msg,
                    relative_points=Decimal(0),
                )
            elif (
                out_form == DataFormat.strict_text
                and san_res.kind == SanitizationResultKind.changed
            ):
                return RelativeSolutionResult(
                    verdict=Verdict.normalization_fail,
                    message=f"Output not normalized. (Check encoding, missing newline at the end or '\\r'.)",
                    relative_points=Decimal(0),
                )

            return self._check()
        elif kind == RunResultKind.RUNTIME_ERROR:
            return RelativeSolutionResult(
                verdict=Verdict.error,
                message=None,
                relative_points=Decimal(0),
            )
        elif kind == RunResultKind.TIMEOUT:
            return RelativeSolutionResult(
                verdict=Verdict.timeout,
                message=None,
                relative_points=Decimal(0),
            )
        else:
            assert_never(kind)

    def message(self, sol_rr: RunResult) -> str:
        """Message about how checking ended."""
        if self.result is None:
            raise RuntimeError(f"Job {self.name} has not finished yet.")

        text = f"input: {self._quote_file_with_path(self.input)}"
        if isinstance(self, RunBatchChecker):
            text += f"correct output: {self._quote_file_with_path(self.correct_output)}"
        text += f"result: {self.result.verdict.name}\n"
        if self.result.message is not None:
            text += f"message: {self.result.message}\n"

        text += "solution:\n"
        text += tab(
            self._format_run_result(
                sol_rr,
                status=sol_rr.kind != RunResultKind.OK,
                stderr_force_content=sol_rr.kind == RunResultKind.RUNTIME_ERROR,
                time=True,
            )
        )
        text += "\n"
        if self.checker_rr is not None:
            text += (
                f"{self.checker_name}:\n"
                + tab(
                    self._format_run_result(
                        self.checker_rr,
                        status=self.checker_rr.stderr_file is None,
                        stderr_force_content=True,
                    )
                )
                + "\n"
            )

        return text

    def verdict_text(self) -> str:
        if self.result is not None:
            if self.result.message is not None:
                return self.result.message
            return self.result.verdict.name
        else:
            return self.state.name

    def verdict_mark(self) -> str:
        if self.state == State.cancelled:
            return "-"
        elif self.result is None:
            return " "
        else:
            return self.result.verdict.mark


class RunBatchChecker(RunChecker):
    """Runs batch checker on single input. (Abstract class)"""

    def __init__(
        self,
        env: Env,
        checker_name: str,
        test: int,
        input_: IInputPath,
        output: IOutputPath,
        correct_output: IOutputPath,
        expected_verdict: Optional[Verdict],
        **kwargs,
    ) -> None:
        super().__init__(
            env=env,
            name=f"Check {output:p}",
            checker_name=checker_name,
            test=test,
            input_=input_,
            checker_log_file=output.to_checker_log(checker_name),
            expected_verdict=expected_verdict,
            **kwargs,
        )
        self.output = output
        self.correct_output_name = correct_output
        self.correct_output = correct_output

    @override
    def _get_solution_run_res_kind(self) -> RunResultKind:
        if "run_solution" in self.prerequisites_results:
            return self.prerequisites_results["run_solution"]
        else:
            # There is no solution (checking static tests against themselves)
            return RunResultKind.OK

    @override
    def _checking_message(self) -> str:
        return (
            f"output {self.output.col(self._env)} for input {self.input.col(self._env)}"
        )



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2026        Daniel Skýpala <skipy@kam.mff.cuni.cz>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from collections import defaultdict
from decimal import Decimal

from pisek.jobs.jobs import State, Job
from pisek.jobs.status import StatusJobManager
from pisek.config.config_types import GenType
from pisek.config.task_config import RunSection
from pisek.env.env import TestingTarget
from pisek.task_jobs.run_result import RunResult
from pisek.task_jobs.task_manager import GENERATE_INPUTS_MAN_CODE, SOLUTION_MAN_CODE
from pisek.task_jobs.solution.solution_result import SolutionResultDetail


class ResourceStatistics(StatusJobManager):
    def __init__(self) -> None:
        super().__init__("Resource statistics")

    def _get_jobs(self) -> list[Job]:
        return []

    def get_status(self) -> str:
        def format_stat(
            run_results: list[RunResult],
            limit: Decimal | int,
            attr: str,
            dec_places: int = 0,
        ) -> list[str]:
            value = max(map(lambda rr: getattr(rr, attr), run_results))
            str_value = f"{value:.{dec_places}f}"

            if limit == 0:
                color = "green"
            elif value < Decimal("0.5") * limit:
                color = "green"
            elif value < Decimal("0.9") * limit:
                color = "yellow"
            elif value <= limit:
                color = "red"
            else:
                color = "white"

            return [self._colored(str_value, color), "/", f"{limit:.{dec_places}f}"]

        def part_statistics(
            name: str, run_results: list[RunResult], run_section: RunSection
        ) -> list[str]:
            return [
                f"{name:<30} ",
                *format_stat(run_results, run_section.time_limit, "time", dec_places=2),
                "s  ",
                *format_stat(run_results, run_section.mem_limit, "memory"),
                "MB  ",
            ]

        if self.state == State.cancelled:
            return self._job_bar(self.name)

        generator_rr: list[RunResult] = self.prerequisites_results[
            GENERATE_INPUTS_MAN_CODE
        ]["generator_run_results"]
        solution_rr: dict[str, list[RunResult]] = defaultdict(list)
        checker_rr: list[RunResult] = []

        for name, data in self.prerequisites_results.items():
            if not name.startswith(SOLUTION_MAN_CODE) or not any(
                data["results"].values()
            ):
                continue

            solution = name[len(SOLUTION_MAN_CODE) :]

            detail: SolutionResultDetail | None
            for detail in data["results"].values():
                if detail is None:
                    continue

                solution_rr[solution].append(detail.solution_run_result)
                if detail.checker_run_result is not None:
                    checker_rr.append(detail.checker_run_result)

        table: list[list[str]] = []

        if (
            self._env.config.tests.in_gen is not None
            and generator_rr
            and self._env.config.tests.gen_type != GenType.cms_old
        ):
            table.append(
                part_statistics(
                    f"generator {self._env.config.tests.in_gen.name}",
                    generator_rr,
                    self._env.config.tests.in_gen,
                )
            )

        if self._env.target in (
            TestingTarget.primary,
            TestingTarget.solutions,
            TestingTarget.all,
        ):
            for name, run_results in solution_rr.items():
                table.append(
                    part_statistics(
                        f"solution {name}",
                        run_results,
                        self._env.config.solutions[name].run,
                    )
                )

            if self._env.config.tests.out_judge is not None:
                table.append(
                    part_statistics(
                        f"judge {self._env.config.tests.out_judge.name}",
                        checker_rr,
                        self._env.config.tests.out_judge,
                    )
                )

        msg = "\n"
        for line in table:
            for part_i in range(len(line)):
                max_len = max(len(l[part_i]) for l in table)
                if part_i == 0:
                    msg += f"{line[part_i]:<{max_len}} "
                else:
                    msg += f"{line[part_i]:>{max_len}} "

            msg += "\n"

        return msg



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2026        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from argparse import ArgumentTypeError
from decimal import Decimal, InvalidOperation


def argparse_Decimal(value: str, type_str: str = "decimal") -> Decimal:
    try:
        dvalue = Decimal(value)
    except InvalidOperation:
        raise ArgumentTypeError(f"invalid {type_str} value: '{value}'")
    if not dvalue.is_finite():
        raise ArgumentTypeError(f"invalid {type_str} value: '{value}'")
    return dvalue


def argparse_positive_Decimal(
    value: str, type_str: str = "positive decimal"
) -> Decimal:
    dvalue = argparse_Decimal(value, type_str)
    if dvalue <= 0:
        raise ArgumentTypeError(f"invalid {type_str} value: '{value}'")
    return dvalue


def argparse_nonnegative_Decimal(
    value: str, type_str: str = "non-negative decimal"
) -> Decimal:
    dvalue = argparse_Decimal(value, type_str)
    if dvalue < 0:
        raise ArgumentTypeError(f"invalid {type_str} value: '{value}'")
    return dvalue



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor, Future, wait

from pisek.env.env import Env
from pisek.jobs.logging import json_logging
from pisek.jobs.jobs import State, Job, JobManager
from pisek.jobs.cache import Cache
from pisek.jobs.reporting import Reporter, CommandLineReporter


class JobPipeline(ABC):
    """Runs given Jobs and JobManagers according to their prerequisites."""

    @abstractmethod
    def __init__(self) -> None:
        self.job_managers: list[JobManager] = []
        self._tmp_lines: int = 0
        self.all_accessed_files: set[str] = set()

    def run_jobs(self, cache: Cache | None, env: Env) -> bool:
        self._reporter: Reporter = CommandLineReporter(env, self.job_managers)

        self._futures: dict[Future, Job] = {}
        self._queue: list[Job] = []
        self._envs: list[Env] = [env.model_copy(deep=True) for _ in range(env.jobs)]

        for job_man in self.job_managers:
            job_man.set_env(env)

        with ThreadPoolExecutor(max_workers=env.jobs) as self._thread_pool:
            self._update(cache, env)
            while not all(man.state.finished() for man in self.job_managers):
                done, _ = wait(
                    self._futures, timeout=0.1, return_when="FIRST_COMPLETED"
                )

                for future in done:
                    exception = future.exception()
                    if exception is not None:
                        raise exception

                    job = self._futures[future]
                    del self._futures[future]
                    self._envs.append(job._env)
                    self._finalize_job(job, cache)

                if not self._update(cache, env):
                    break

                self._reporter.update(list(self._futures.values()))

            for job in self._futures.values():
                if job.state == State.succeeded:
                    self._finalize_job(job, cache)
                else:
                    job.cancel()
            self._reporter.update([])

        if cache is not None:
            cache.export()  # Save last version of cache
        json_logging.write()

        return any(man.state == State.failed for man in self.job_managers)

    def _update(self, cache: Cache | None, env: Env) -> bool:
        """Updates currently running JobManagers and Jobs, runs new ones
        and returns if we should continue in testing."""

        # Start new managers
        for manager in self.job_managers:
            if manager.state == State.in_queue and manager.prerequisites == 0:
                self._queue.extend(manager.create_jobs())
                if manager.any_failed():
                    manager.finalize()
                    self._reporter.report_manager(manager)
                    if not env.full:
                        return False
                break  # We don't want to start many managers at once because that can lead to UI freeze

        # Process new jobs
        new_queue: list[Job] = []
        to_run: list[tuple[Job, Env]] = []
        for job in self._queue:
            if len(self._envs) > 0 and job.prerequisites == 0:
                job.prepare(cache)
                if job.state.finished():
                    self._finalize_job(job, cache)
                elif job.state:
                    to_run.append((job, self._envs.pop()))
            else:
                new_queue.append(job)

        self._queue = new_queue

        # Update managers
        for manager in self.job_managers:
            if manager.state == State.running:
                manager.update()
                if manager.ready() or manager.any_failed():
                    manager.finalize()
                    self._reporter.report_manager(manager)
                if manager.any_failed() and not env.full:
                    return False

        # Start new jobs
        for job, env in to_run:
            if job.state == State.in_queue:
                self._futures[self._thread_pool.submit(job.run, env)] = job
            else:
                self._envs.append(env)

        return True

    def _finalize_job(self, job: Job, cache: Cache | None) -> None:
        job.finalize(cache)
        self.all_accessed_files |= job.accessed_files
        self._reporter.report_job(job)



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <skipy@kam.mff.cuni.cz>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from contextlib import ExitStack
from dataclasses import dataclass, field
from decimal import Decimal
from math import ceil
import os
from tempfile import NamedTemporaryFile, TemporaryDirectory
from typing import Optional, Any, Union, Callable
import signal
import subprocess
from tempfile import NamedTemporaryFile, TemporaryDirectory
from typing import assert_never, Optional, Any, Union, Callable

from pisek.config.task_config import ProgramRole, RunSection
from pisek.env.env import Env
from pisek.utils.paths import TaskPath, LogPath
from pisek.jobs.jobs import PipelineItemFailure
from pisek.utils.text import tab
from pisek.task_jobs.run_result import RunResultKind, RunResult
from pisek.task_jobs.task_job import TaskJob


@dataclass
class ProgramPoolItem:
    name: str
    executable: TaskPath
    args: list[str]
    time_limit: Decimal
    clock_limit: Decimal
    mem_limit: int
    process_limit: int
    stdin: Optional[Union[TaskPath, int]]
    stdout: Optional[Union[TaskPath, int]]
    stderr: Optional[TaskPath]
    env: dict[str, str] = field(default_factory=lambda: {})

    def to_popen(self, minibox: str, meta_file: str) -> dict[str, Any]:
        """Returns subprocess.Popen args for executing this PoolItem."""
        result: dict[str, Any] = {}

        minibox_args = []
        minibox_args.append(f"--time={self.time_limit}")
        minibox_args.append(f"--wall-time={self.clock_limit}")
        minibox_args.append(f"--mem={self.mem_limit*1024}")
        # Minibox doesn't support limits for multiple processes (#613)
        if self.process_limit != 1:
            minibox_args.append(f"--processes=0")

        for std in ("stdin", "stdout", "stderr"):
            attr = getattr(self, std)
            if isinstance(attr, TaskPath):
                minibox_args.append(f"--{std}={attr.abspath}")
            elif getattr(self, std) is None:
                minibox_args.append(f"--{std}=/dev/null")

            if isinstance(attr, int):
                result[std] = attr
            else:
                result[std] = subprocess.PIPE

        for key, val in self.env.items():
            minibox_args.append(f"--env={key}={val}")

        minibox_args.append("--silent")
        minibox_args.append(f"--meta={meta_file}")

        result["args"] = (
            [minibox]
            + minibox_args
            + ["--run", "--", self.executable.abspath]
            + self.args
        )
        return result


class ProgramsJob(TaskJob):
    """Job that deals with a program."""

    def __init__(self, env: Env, name: str, **kwargs) -> None:
        super().__init__(env=env, name=name, **kwargs)
        self._program_pool: list[ProgramPoolItem] = []
        self._callback: Optional[Callable[[subprocess.Popen], None]] = None

    @staticmethod
    def _env_disjoint_union(
        env1: dict[str, str], env2: dict[str, str]
    ) -> dict[str, str]:
        union = env1 | env2
        if len(union) != len(env1) + len(env2):
            duplicates = list(set(env1.keys()) & set(env2.keys()))
            raise PipelineItemFailure(
                f"Environment variable '{duplicates[0]}' is used by the protocol and cannot be overwritten."
            )
        return union

    def _load_executable(
        self,
        path: TaskPath,
        args: list[str],
        time_limit: Decimal,
        clock_limit: Decimal,
        mem_limit: int,
        process_limit: int,
        stdin: Optional[Union[TaskPath, int]] = None,
        stdout: Optional[Union[TaskPath, int]] = None,
        stderr: Optional[LogPath] = None,
        env: dict[str, str] = {},
    ):
        if self._is_file(path):
            self._access_file(path)
            executable = path
        elif self._is_dir(path):
            self._access_dir(path)
            executable = path.join("run")

        if self._is_dir(executable):
            raise PipelineItemFailure(
                f"Cannot execute '{executable:p}': Is a directory."
            )
        elif not self._exists(executable):
            raise PipelineItemFailure(f"Cannot execute '{executable:p}': File missing.")

        if isinstance(stdin, TaskPath):
            self._access_file(stdin)
        if isinstance(stdout, TaskPath):
            self.make_filedirs(stdout)
            self._access_file(stdout)
        if isinstance(stderr, TaskPath):
            self.make_filedirs(stderr)
            self._access_file(stderr)

        self._program_pool.append(
            ProgramPoolItem(
                name=path.name,
                executable=executable,
                args=args,
                time_limit=time_limit,
                clock_limit=clock_limit,
                mem_limit=mem_limit,
                process_limit=process_limit,
                stdin=stdin,
                stdout=stdout,
                stderr=stderr,
                env=env,
            )
        )

    def _load_program(
        self,
        program_role: ProgramRole,
        program: RunSection,
        args: list[str] = [],
        stdin: Optional[Union[TaskPath, int]] = None,
        stdout: Optional[Union[TaskPath, int]] = None,
        stderr: Optional[LogPath] = None,
        env={},
    ) -> None:
        """Adds program to execution pool."""
        time_limit: Decimal | None = None
        if program_role.is_solution():
            time_limit = self._env.time_limit

        self._load_executable(
            path=program.executable,
            args=program.args + args,
            time_limit=program.time_limit if time_limit is None else time_limit,
            clock_limit=program.clock_limit(time_limit),
            mem_limit=program.mem_limit,
            process_limit=program.process_limit,
            stdin=stdin,
            stdout=stdout,
            stderr=stderr,
            env=self._env_disjoint_union(env, program.env),
        )

    def _load_callback(self, callback: Callable[[subprocess.Popen], None]) -> None:
        if self._callback is not None:
            raise RuntimeError("Callback already loaded.")
        self._callback = callback

    def _read_run_result(
        self, pool_item: ProgramPoolItem, process: subprocess.Popen, meta_file: str
    ) -> RunResult:
        assert process.stderr is not None  # To make mypy happy

        if process.returncode not in (0, 1):
            raise PipelineItemFailure(
                f"Minibox error:\n{tab(process.stderr.read().decode())}"
            )

        with open(meta_file) as f:
            meta_raw = f.read().strip().split("\n")
        meta = {key: val for key, val in map(lambda x: x.split(":", 1), meta_raw)}

        time = Decimal(meta["time"])
        wall_time = Decimal(meta["time-wall"])
        memory = ceil(int(meta["max-rss"]) / 1024)

        if process.returncode == 0:
            return RunResult(
                RunResultKind.OK,
                0,
                time,
                wall_time,
                memory,
                pool_item.stdin,
                pool_item.stdout,
                pool_item.stderr,
                "Exited with return code 0",
            )

        elif process.returncode == 1:
            if meta["status"] in ("RE", "SG"):
                if meta["status"] == "RE":
                    return_code = int(meta["exitcode"])
                elif meta["status"] == "SG":
                    return_code = int(meta["exitsig"])
                    meta["message"] += f" ({signal.Signals(return_code).name})"

                return RunResult(
                    RunResultKind.RUNTIME_ERROR,
                    return_code,
                    time,
                    wall_time,
                    memory,
                    pool_item.stdin,
                    pool_item.stdout,
                    pool_item.stderr,
                    meta["message"],
                )

            elif meta["status"] == "TO":
                time_limit = (
                    f"{pool_item.time_limit}s"
                    if time > pool_item.time_limit
                    else f"{pool_item.clock_limit}ws"
                )
                return RunResult(
                    RunResultKind.TIMEOUT,
                    -1,
                    time,
                    wall_time,
                    memory,
                    pool_item.stdin,
                    pool_item.stdout,
                    pool_item.stderr,
                    f"Timeout after {time_limit}",
                )
            else:
                raise RuntimeError(f"Unknown minibox status {meta['message']}.")

        assert False

    def _run_programs(self) -> list[RunResult]:
        """Runs all programs in execution pool."""
        running_pool: list[subprocess.Popen] = []
        meta_files: list[str] = []
        tmp_dirs: list[TemporaryDirectory] = []
        minibox = TaskPath.executable_path("_minibox").abspath
        with ExitStack() as exit_stack:
            for pool_item in self._program_pool:
                meta_file = NamedTemporaryFile(prefix="pisek_", delete_on_close=False)
                exit_stack.enter_context(meta_file)
                meta_file.close()
                meta_files.append(meta_file.name)

                popen = pool_item.to_popen(minibox, meta_file.name)
                self._log(
                    "debug",
                    "Executing '" + " ".join(popen["args"]) + "'",
                    bypass_cache=True,
                )

                tmp_dir = TemporaryDirectory(prefix="pisek_")
                exit_stack.enter_context(tmp_dir)
                tmp_dirs.append(tmp_dir)

                running_pool.append(subprocess.Popen(**popen, cwd=tmp_dir.name))

            if self._callback is not None:
                callback_exec = False
                while True:
                    states = [process.poll() is not None for process in running_pool]
                    if not callback_exec and any(states):
                        callback_exec = True
                        self._callback(running_pool[states.index(True)])

                    if all(states):
                        break

            run_results = []
            for pool_item, process, tmp_dir, meta_file_path in zip(
                self._program_pool, running_pool, tmp_dirs, meta_files
            ):
                self._wait_for_subprocess(process)
                run_results.append(
                    self._read_run_result(pool_item, process, meta_file_path)
                )

                created_paths = self._listdir(TaskPath(tmp_dir.name))
                if created_paths:
                    file_list = "".join(
                        [
                            (
                                self._quote_file_with_name(p)
                                if self._is_file(p)
                                else p.col_name(self._env) + "\n"
                            )
                            for p in created_paths
                        ]
                    ).removesuffix("\n")
                    raise PipelineItemFailure(
                        f"'{pool_item.name}' created paths in cwd:\n" + tab(file_list)
                    )

        return run_results

    def _run_program(
        self,
        program_role: ProgramRole,
        program: RunSection,
        **kwargs,
    ) -> RunResult:
        """Loads one program and runs it."""
        self._load_program(program_role, program, **kwargs)
        return self._run_programs()[0]

    def _run_tool(
        self,
        program: str,
        args: list[str] = [],
        stdin: Optional[Union[TaskPath, int]] = None,
        stdout: Optional[Union[TaskPath, int]] = None,
        stderr: Optional[LogPath] = None,
        env={},
    ) -> RunResult:
        """Loads one program and runs it."""
        self._load_executable(
            TaskPath.executable_path(program),
            args=args,
            time_limit=Decimal(300),
            clock_limit=Decimal(300),
            mem_limit=0,
            process_limit=1,
            stdin=stdin,
            stdout=stdout,
            stderr=stderr,
            env=env,
        )
        return self._run_programs()[0]



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2025        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass, asdict
import json
import logging
import os
import threading
from typing import Callable, Literal

from pisek.utils.text import fatal_user_error
from pisek.utils.colors import remove_colors
from pisek.utils.paths import INTERNALS_DIR

root_logger = logging.getLogger()
logger = logging.getLogger(__name__)

LOG_JSON_FILE = "log.json"

LogLevel = (
    Literal["critical"]
    | Literal["error"]
    | Literal["warning"]
    | Literal["info"]
    | Literal["debug"]
)


@dataclass
class LogEntry:
    scope: str
    level: LogLevel
    message: str

    def __init__(self, scope: str, level: LogLevel, message: str) -> None:
        self.scope = scope
        self.level = level
        self.message = remove_colors(message)


def log(log_entry: LogEntry) -> None:
    getattr(root_logger, log_entry.level)(
        log_entry.message, extra={"scope": log_entry.scope}
    )


def map_log_level(log_level: str) -> int:
    log_level = log_level.lower()
    if log_level == "debug":
        return logging.DEBUG
    elif log_level == "info":
        return logging.INFO
    elif log_level == "warning":
        return logging.WARNING
    elif log_level == "error":
        return logging.ERROR
    elif log_level == "critical":
        return logging.CRITICAL
    else:
        fatal_user_error(f"Invalid log level: '{log_level}'")


class _CallbackHandler(logging.Handler):
    def __init__(self, callback: Callable[[LogEntry], None], level=logging.NOTSET):
        self.__callback = callback
        super().__init__(level)

    def emit(self, record):
        try:
            if hasattr(record, "scope"):
                scope = record.scope
            else:
                scope = "unknown"
            self.__callback(LogEntry(scope, record.levelname.lower(), record.msg))
        except Exception:
            self.handleError(record)


class _JSONLogging:
    def __init__(self) -> None:
        self._lock = threading.Lock()
        self._enabled = False
        self._entries: list[LogEntry] = []

    def enable(self) -> None:
        self._enabled = True

    def get_handler(self) -> logging.Handler:
        return _CallbackHandler(self._log)

    def _log(self, entry: LogEntry) -> None:
        self._entries.append(entry)

    def write(self) -> None:
        if not self._enabled:
            return

        os.makedirs(INTERNALS_DIR, exist_ok=True)
        with open(os.path.join(INTERNALS_DIR, LOG_JSON_FILE), "w") as f:
            json.dump(list(map(asdict, self._entries)), f)


json_logging = _JSONLogging()



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import glob
import os
import shutil
from typing import Iterable

from pisek.utils.paths import BUILD_DIR, TESTS_DIR, INTERNALS_DIR, TaskPath


class ChangedCWD:
    def __init__(self, path):
        self._path = path

    def __enter__(self):
        self._orig_path = os.getcwd()
        os.chdir(self._path)

    def __exit__(self, exc_type, exc_value, exc_traceback):
        os.chdir(self._orig_path)


def clean_task_dir(task_dir: str) -> None:
    for subdir in [BUILD_DIR, TESTS_DIR, INTERNALS_DIR]:
        full = os.path.join(task_dir, subdir)
        try:
            shutil.rmtree(full)
        except FileNotFoundError:
            pass


def clean_non_relevant_files(accessed_files: set[str]) -> None:
    accessed_dirs = {os.path.dirname(file) for file in accessed_files}
    for root, _, files in os.walk(TESTS_DIR):
        for file in files:
            path = os.path.join(root, file)
            if root in accessed_dirs and path not in accessed_files:
                os.remove(path)


def globs_to_files(
    globs: Iterable[str], directory: TaskPath, exclude: Iterable[str] = ()
) -> list[TaskPath]:
    files_per_glob = [
        glob.glob(g, root_dir=directory.path, recursive=True, include_hidden=True)
        for g in globs
    ]
    files = sorted(set(sum(files_per_glob, start=[])))
    task_paths = [TaskPath.from_abspath(directory.path, file) for file in files]
    exclude_tp = [TaskPath.from_abspath(directory.path, path) for path in exclude]
    return sorted(
        tp for tp in task_paths if all(not exc_p.is_prefix(tp) for exc_p in exclude_tp)
    )
